from bedrock_protocol.packets.minecraft_packet_ids import MinecraftPacketIds
from bedrock_protocol.packets.packet.packet_base import Packet
from bstream import BinaryStream, ReadOnlyBinaryStream


class NetworkStackLatencyPacket(Packet):
    timestamp: int
    from_server: bool

    def __init__(self, timestamp: int = 0, from_server: bool = False):
        super().__init__()
        self.timestamp = timestamp
        self.from_server = from_server

    def get_packet_id(self) -> MinecraftPacketIds:
        return MinecraftPacketIds.Ping

    def get_packet_name(self) -> str:
        return "NetworkStackLatencyPacket"

    def write(self, stream: BinaryStream) -> None:
        stream.write_unsigned_int64(self.timestamp)
        stream.write_bool(self.from_server)

    def read(self, stream: ReadOnlyBinaryStream) -> None:
        self.timestamp = stream.get_unsigned_int64()
        self.from_server = stream.get_bool()



from abc import ABC, abstractmethod

from endstone import Player


class Graphic(ABC):

    @abstractmethod
    def send(self, player: Player):
        pass

    @abstractmethod
    def send_data(self, player: Player):
        pass

    @abstractmethod
    def open(self, player: Player):
        pass

    @abstractmethod
    def remove(self, player: Player):
        pass



from typing import Optional, Callable, TYPE_CHECKING

from endstone import Player
from endstone._internal.endstone_python import ItemStack

from endstone_inventoryui.manager.player_manager import find_session, create_session
from endstone_inventoryui.menu.inventory import UIInventory
from endstone_inventoryui.menu.menu_type import MenuType

if TYPE_CHECKING:
    from endstone_inventoryui.manager import Session


class Menu:

    def __init__(self, type: MenuType, name: str = ""):
        self._name = name
        self._type = type
        self._inventory: UIInventory = UIInventory(type.container_size,
                                                   slot_updated=self._on_slot_changed)
        self._listener: Optional[Callable[[Player, int, ItemStack, UIInventory], None]] = None
        self._open_listener: Optional[Callable[[Player], None]] = None
        self._close_listener: Optional[Callable[[Player], None]] = None
        self._sessions: set['Session'] = set()

    @property
    def inventory(self):
        return self._inventory

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    def set_name(self, name: str):
        """
        Set the display name of the menu.

        Args:
            name: The new display name to show at the top of the menu.
        """
        self._name = name

    def set_listener(self, listener: Callable[[Player, int, ItemStack, UIInventory], None]):
        self._listener = listener

    def set_open_listener(self, listener: Callable[[Player], None]):
        """
        Set the callback for menu open events.

        Args:
            listener: A function called when a player opens this menu.
        """

        self._open_listener = listener

    def set_close_listener(self, listener: Callable[[Player], None]):
        """
        Set the callback for menu close events.

        Args:
            listener: A function called when a player closes this menu.
        """
        self._close_listener = listener

    def send_to(self, player: Player):
        """
        Display this menu to a player.

        If the player already has a menu open, this menu is queued and will
        be shown after the current menu is closed. Otherwise, a new session
        is created and the menu is displayed immediately.

        Args:
            player: The player to show the menu to.
        """
        existing_session = find_session(player)
        if existing_session is not None:
            existing_session.pending.append(self)
        else:
            session = create_session(player)
            session.menu = self
            session.send_menu()

    def close(self, player: Player) -> bool:
        """
        Close this specific menu for a player.

        Only closes the menu if it matches the player's currently open menu
        and the session is not already in a closing state.
        """
        from endstone_inventoryui.manager import Session
        session = find_session(player)
        if session is not None:
            if session.menu == self and session.state != Session.State.CLOSING:
                session.close()
                return True
        return False

    def close_all(self) -> None:
        """
        Close this menu for all players currently viewing it.
        """
        from endstone_inventoryui.manager import Session
        for s in self._sessions:
            if s.state != Session.State.CLOSING:
                s.close()

    def get_viewers(self) -> list[Player]:
        """
        Get a list of players who currently have this menu open.
        """
        from endstone_inventoryui.manager import Session
        players = []
        for s in self._sessions:
            if s.state == Session.State.OPEN:
                players.append(s.player)
        return players

    def _on_slot_changed(self, slot: int) -> None:
        from endstone_inventoryui.manager import Session
        for session in self._sessions:
            if session.state == Session.State.OPEN:
                session.update_slot(slot)

    def _add_session(self, session: 'Session') -> None:
        self._sessions.add(session)

    def _remove_session(self, session: 'Session') -> None:
        self._sessions.discard(session)



from .main import InventoryUIPlugin
from .menu.inventory import UIInventory
from .menu.menu import Menu
from .menu.menu_type import MenuType

__all__ = [
    "InventoryUIPlugin",
    "Menu",
    "MenuType",
    "UIInventory",
]



import math
import random

from bedrock_protocol.packets.packet import UpdateBlockPacket
from bedrock_protocol.packets.types import BlockPos, NetworkBlockPosition
from endstone import Player, Server

from endstone_inventoryui.network.network_stack_latency_packet import NetworkStackLatencyPacket

server: Server | None = None


def send_ack_packet(player: Player) -> int:
    """
    Sends a ping packet the client is expected to respond to

    :param player: Player
    :returns: Timestamp value expected to be sent by the player
    """
    timestamp = random.randint(1, 32767)
    pk = NetworkStackLatencyPacket(timestamp, True)
    player.send_packet(pk.get_packet_id(), pk.serialize())
    return timestamp * 1_000_000


def send_block(player: Player, name: str, pos: BlockPos):
    block_data = player.server.create_block_data(name)
    pk = UpdateBlockPacket(NetworkBlockPosition(pos.x, pos.y, pos.z), block_data.runtime_id, 0b0010, 0)
    player.send_packet(pk.get_packet_id(), pk.serialize())


def get_block_behind(player: Player, distance: int = 1) -> BlockPos:
    location = player.location
    yaw = location.yaw

    behind_yaw = yaw + 180
    yaw_rad = math.radians(behind_yaw)
    x = -math.sin(yaw_rad) * distance
    z = math.cos(yaw_rad) * distance

    behind_x = location.x + x
    behind_y = location.y + 1
    behind_z = location.z + z

    return BlockPos(math.floor(behind_x), math.floor(behind_y), math.floor(behind_z))


def west(pos: BlockPos) -> BlockPos:
    return BlockPos(pos.x - 1, pos.y, pos.z)


def east(pos: BlockPos) -> BlockPos:
    return BlockPos(pos.x + 1, pos.y, pos.z)



from typing import overload, Callable

from endstone._internal.endstone_python import ItemStack

from endstone_inventoryui.util.item_utils import clone_item


class UIInventory:
    """
    Inventory used for menus

    Based on Endstone's inventory interface
    """

    def __init__(self, size: int = 36, max_stack_size: int = 64,
                 slot_updated: Callable[[int], None] | None = None):
        self._size = size
        self._max_stack_size = max_stack_size
        self._slots: list[ItemStack | None] = [None] * size
        self._slot_updated = slot_updated

    def _notify_slot_update(self, index: int) -> None:
        """Calls the slot_updated callback if one is registered."""
        if self._slot_updated is not None:
            self._slot_updated(index)

    @property
    def size(self) -> int:
        """Returns the size of the inventory"""
        return self._size

    @property
    def max_stack_size(self) -> int:
        """Returns the maximum stack size for an ItemStack in this inventory."""
        return self._max_stack_size

    def get_item(self, index: int) -> ItemStack:
        """Returns the ItemStack found in the slot at the given index"""
        if not 0 <= index < self._size:
            raise IndexError(f"Slot index {index} out of range")
        item = self._slots[index]
        if item is None:
            return ItemStack("minecraft:air")
        return item

    def set_item(self, index: int, item: ItemStack) -> None:
        """Stores the ItemStack at the given index of the inventory."""
        if not 0 <= index < self._size:
            raise IndexError(f"Slot index {index} out of range")
        self._slots[index] = item
        self._notify_slot_update(index)

    def add_item(self, *args: ItemStack) -> dict[int, ItemStack]:
        """
        Stores the given ItemStacks in the inventory.
        Tries to fill existing stacks and empty slots.
        Returns what couldn't be stored.
        """
        leftover = {}
        for i, item in enumerate(args):
            if item is None or item.amount <= 0:
                continue
            remaining = clone_item(item)

            for slot_idx in range(self._size):
                if remaining.amount <= 0:
                    break
                slot_item = self._slots[slot_idx]
                if slot_item is not None and slot_item.is_similar(remaining):
                    max_add = min(remaining.amount,
                                  min(self._max_stack_size, slot_item.max_stack_size) - slot_item.amount)
                    if max_add > 0:
                        slot_item.amount += max_add
                        remaining.amount -= max_add
                        self._notify_slot_update(slot_idx)

            for slot_idx in range(self._size):
                if remaining.amount <= 0:
                    break
                if self._slots[slot_idx] is None:
                    max_add = min(remaining.amount,
                                  min(self._max_stack_size, remaining.max_stack_size))
                    new_item = clone_item(remaining)
                    new_item.amount = max_add
                    self._slots[slot_idx] = new_item
                    remaining.amount -= max_add
                    self._notify_slot_update(slot_idx)
            if remaining.amount > 0:
                leftover[i] = remaining
        return leftover

    def remove_item(self, *args: ItemStack) -> dict[int, ItemStack]:
        """
        Removes the given ItemStacks from the inventory.
        Returns what couldn't be removed.
        """
        leftover = {}
        for i, item in enumerate(args):
            if item is None or item.amount <= 0:
                continue
            to_remove = item.amount
            for slot_idx in range(self._size):
                if to_remove <= 0:
                    break
                slot_item = self._slots[slot_idx]
                if slot_item is not None and slot_item.is_similar(item):
                    removed = min(to_remove, slot_item.amount)
                    slot_item.amount -= removed
                    to_remove -= removed
                    if slot_item.amount <= 0:
                        self._slots[slot_idx] = None
                    self._notify_slot_update(slot_idx)
            if to_remove > 0:
                leftover_item = clone_item(item)
                leftover_item.amount = to_remove
                leftover[i] = leftover_item
        return leftover

    @property
    def contents(self) -> list[ItemStack]:
        """Returns all ItemStacks from the inventory"""
        return self._slots.copy()

    @contents.setter
    def contents(self, items: list[ItemStack]) -> None:
        for i in range(self._size):
            old_item = self._slots[i]
            new_item = items[i] if i < len(items) else None
            self._slots[i] = new_item
            if old_item != new_item:
                self._notify_slot_update(i)

    @overload
    def contains(self, item: ItemStack, amount: int) -> bool:
        ...

    @overload
    def contains(self, item: ItemStack) -> bool:
        ...

    @overload
    def contains(self, type: str) -> bool:
        ...

    def contains(self, item, amount: int = None) -> bool:
        if isinstance(item, str):
            # Check by type
            return any(s is not None and s.type == item for s in self._slots)
        elif isinstance(item, ItemStack):
            if amount is not None:
                # Count exact matches
                count = sum(1 for s in self._slots if s is not None and s == item)
                return count >= amount
            else:
                return any(s is not None and s == item for s in self._slots)
        return False

    @overload
    def contains_at_least(self, item: ItemStack, amount: int) -> bool:
        ...

    @overload
    def contains_at_least(self, type: str, amount: int) -> bool:
        ...

    def contains_at_least(self, item, amount: int) -> bool:
        if isinstance(item, str):
            total = sum(s.amount for s in self._slots if s is not None and s.type == item)
            return total >= amount
        elif isinstance(item, ItemStack):
            total = sum(s.amount for s in self._slots if s is not None and s.is_similar(item))
            return total >= amount
        return False

    @overload
    def all(self, item: ItemStack) -> dict[int, ItemStack]:
        ...

    @overload
    def all(self, type: str) -> dict[int, ItemStack]:
        ...

    def all(self, item) -> dict[int, ItemStack]:
        result = {}
        if isinstance(item, str):
            for i, s in enumerate(self._slots):
                if s is not None and s.type == item:
                    result[i] = s
        elif isinstance(item, ItemStack):
            for i, s in enumerate(self._slots):
                if s is not None and s == item:
                    result[i] = s
        return result

    @overload
    def first(self, item: ItemStack) -> int:
        ...

    @overload
    def first(self, type: str) -> int:
        ...

    def first(self, item) -> int:
        if isinstance(item, str):
            for i, s in enumerate(self._slots):
                if s is not None and s.type == item:
                    return i
        elif isinstance(item, ItemStack):
            for i, s in enumerate(self._slots):
                if s is not None and s == item:
                    return i
        return -1

    @property
    def first_empty(self) -> int:
        """Returns the first empty Slot."""
        for i, s in enumerate(self._slots):
            if s is None:
                return i
        return -1

    @property
    def is_empty(self) -> bool:
        """Check whether this inventory is empty."""
        return all(s is None for s in self._slots)

    @overload
    def remove(self, item: ItemStack) -> None:
        ...

    @overload
    def remove(self, type: str) -> None:
        ...

    def remove(self, item) -> None:
        if isinstance(item, str):
            for i, s in enumerate(self._slots):
                if s is not None and s.type == item:
                    self._slots[i] = None
                    self._notify_slot_update(i)
        elif isinstance(item, ItemStack):
            for i, s in enumerate(self._slots):
                if s is not None and s == item:
                    self._slots[i] = None
                    self._notify_slot_update(i)

    @overload
    def clear(self, index: int) -> None:
        ...

    @overload
    def clear(self) -> None:
        ...

    def clear(self, index: int = None) -> None:
        if index is not None:
            if 0 <= index < self._size:
                if self._slots[index] is not None:
                    self._slots[index] = None
                    self._notify_slot_update(index)
        else:
            for i in range(self._size):
                if self._slots[i] is not None:
                    self._slots[i] = None
                    self._notify_slot_update(i)

    def __len__(self) -> int:
        return self._size

    def __getitem__(self, index: int) -> ItemStack:
        return self.get_item(index)

    def __setitem__(self, index: int, item: ItemStack) -> None:
        self.set_item(index, item)

    def __contains__(self, item) -> bool:
        return self.contains(item)



from dataclasses import dataclass
from enum import Enum


class MenuType(Enum):
    @dataclass(frozen=True)
    class Value:
        is_pair: bool
        block_id: str
        container_type: int
        block_actor_id: str
        container_size: int

    CHEST = Value(
        is_pair=False,
        block_id="minecraft:chest",
        container_type=0x0,
        block_actor_id="Chest",
        container_size=27
    )
    DOUBLE_CHEST = Value(
        is_pair=True,
        block_id="minecraft:chest",
        container_type=0x0,
        block_actor_id="Chest",
        container_size=54
    )
    HOPPER = Value(
        is_pair=False,
        block_id="minecraft:hopper",
        container_type=0x8,
        block_actor_id="Hopper",
        container_size=5
    )
    DISPENSER = Value(
        is_pair=False,
        block_id="minecraft:dispenser",
        container_type=0x6,
        block_actor_id="Dispenser",
        container_size=9
    )

    @property
    def is_pair(self) -> bool:
        return self.value.is_pair

    @property
    def block_id(self) -> str:
        """Get the Minecraft block ID for this menu type."""
        return self.value.block_id

    @property
    def container_type(self) -> int:
        """Get the container type ID for this menu type."""
        return self.value.container_type

    @property
    def block_actor_id(self) -> str:
        """Get the block actor ID for this menu type."""
        return self.value.block_actor_id

    @property
    def container_size(self) -> int:
        """Get the container size (number of slots) for this menu type."""
        return self.value.container_size



from bedrock_protocol.packets import MinecraftPacketIds
from bedrock_protocol.packets.enums import ItemStackRequestActionType
from bedrock_protocol.packets.packet import ContainerClosePacket, ItemRegistryPacket, ItemStackRequestPacket
from endstone._internal.endstone_python import PacketSendEvent
from endstone.event import event_handler, EventPriority, PlayerQuitEvent, PacketReceiveEvent
from endstone.plugin import Plugin

from .manager import Session
from .manager.player_manager import find_session, close_session
from .network.network_stack_latency_packet import NetworkStackLatencyPacket
from .util.item_utils import all_item_data, add_item_data


class EventListener:
    def __init__(self, plugin: Plugin):
        self._plugin = plugin

    @event_handler(priority=EventPriority.NORMAL)
    def on_packet_receive(self, event: PacketReceiveEvent):
        packet_id = event.packet_id
        player = event.player
        if player is None:
            return
        if packet_id == MinecraftPacketIds.Ping:
            session = find_session(player)
            if session is not None:
                pk = NetworkStackLatencyPacket()
                pk.deserialize(event.payload)
                if session.ack_timestamp == pk.timestamp:
                    match session.state:
                        case Session.State.GRAPHIC_SENT:
                            session.update_state(Session.State.GRAPHIC_RECEIVED)
                        case Session.State.GRAPHIC_DATA_SENT:
                            session.update_state(Session.State.GRAPHIC_DATA_RECEIVED)
                        case Session.State.OPENING:
                            if session.open_attempts >= Session.MAX_OPEN_ATTEMPTS:
                                # close session after max open attempts
                                session.close()
                                return
                            session.open_attempts += 1
                            session.open()

        elif packet_id == MinecraftPacketIds.ContainerClose:
            session = find_session(player)
            if session is not None:
                pk = ContainerClosePacket()
                pk.deserialize(event.payload)
                # if pk.container_id == 0xFF and pk.container_type == 0xF7 and pk.is_server_side == False:
                #    if session.state != Session.State.OPENING:
                #        return
                #    if session.open_attempts >= Session.MAX_OPEN_ATTEMPTS:
                #        return
                #    session.open_attempts += 1
                #    session.open()
                #    return

                if pk.container_id == Session.CONTAINER_ID:
                    if session.menu._close_listener is not None:
                        session.menu._close_listener(player)
                    if session.pending:
                        if session.state != Session.State.CLOSING:
                            session.close()
                        session.menu = session.pending.popleft()
                        session.send_menu()
                    else:
                        close_session(player)

        elif packet_id == MinecraftPacketIds.PacketViolationWarning:
            session = find_session(player)
            if session is not None:
                if session.state == Session.State.OPENING:
                    session.update_state(Session.State.OPEN)
                    if session.menu._open_listener is not None:
                        session.menu._open_listener(player)

        elif packet_id == MinecraftPacketIds.ItemStackRequest:
            session = find_session(player)
            if session is not None and session.state == Session.State.OPEN:
                menu = session.menu
                pk = ItemStackRequestPacket()
                pk.deserialize(event.payload)
                for req_data in pk.request.request_data:
                    for action in req_data.request_actions:
                        if action.action_type in (ItemStackRequestActionType.Take, ItemStackRequestActionType.Place):
                            slot = action.action_data.source.slot
                            source_id = action.action_data.source.container.container_enum
                            if source_id != 7:  # LEVEL_ENTITY
                                return
                            if menu._listener is not None:
                                item_clicked = menu.inventory.get_item(slot)
                                menu._listener(player, slot, item_clicked, menu.inventory)
                                return

    @event_handler
    def on_packet_send(self, event: PacketSendEvent):
        if event.packet_id == MinecraftPacketIds.ItemRegistryPacket and len(all_item_data()) == 0:
            pk = ItemRegistryPacket()
            pk.deserialize(event.payload)
            for item in pk.item_registry:
                add_item_data(item.item_name, item)

    @event_handler
    def on_player_quit(self, event: PlayerQuitEvent):
        close_session(event.player)



from bedrock_protocol.packets.packet import BlockActorDataPacket, ContainerOpenPacket
from bedrock_protocol.packets.types import BlockPos, NetworkBlockPosition
from endstone import Player
from rapidnbt import CompoundTag

from endstone_inventoryui.menu import Menu
from endstone_inventoryui.menu.graphic.graphic import Graphic
from endstone_inventoryui.util.utils import send_block


class BlockGraphic(Graphic):
    menu: Menu
    pos: BlockPos

    def __init__(self, menu: Menu, pos: BlockPos):
        self.menu = menu
        self.pos = pos

    def send(self, player: Player):
        send_block(player, self.menu.type.block_id, self.pos)

    def send_data(self, player: Player):
        tag = CompoundTag()
        pos = self.pos
        tag.set("id", self.menu.type.block_actor_id)
        tag.set("CustomName", self.menu.name)
        pk = BlockActorDataPacket(NetworkBlockPosition(pos.x, pos.y, pos.z), tag)
        player.send_packet(pk.get_packet_id(), pk.serialize())

    def open(self, player: Player):
        from endstone_inventoryui.manager import Session
        pos = NetworkBlockPosition(self.pos.x, self.pos.y, self.pos.z)
        pk = ContainerOpenPacket(Session.CONTAINER_ID, self.menu.type.container_type, pos)
        player.send_packet(pk.get_packet_id(), pk.serialize())

    def remove(self, player: Player):
        block = player.dimension.get_block_at(self.pos.x, self.pos.y, self.pos.z)
        send_block(player, block.type, self.pos)



from bedrock_protocol.packets.minecraft_packet_ids import MinecraftPacketIds
from bedrock_protocol.packets.packet.packet_base import Packet
from bedrock_protocol.packets.types import FullContainerName
from bstream import BinaryStream, ReadOnlyBinaryStream

from endstone_inventoryui.network.item_stack_wrapper import ItemStackWrapper


class InventorySlotPacket(Packet):

    def __init__(self,
                 container_id: int = 0,
                 slot: int = 0,
                 container_name: FullContainerName | None = None,
                 storage: ItemStackWrapper | None = None,
                 item: ItemStackWrapper | None = None
                 ):
        super().__init__()
        self.container_id = container_id
        self.slot = slot
        self.container_name = container_name or FullContainerName()
        self.storage = storage or ItemStackWrapper()
        self.item = item or ItemStackWrapper()

    def get_packet_id(self) -> MinecraftPacketIds:
        return MinecraftPacketIds.InventorySlot

    def get_packet_name(self) -> str:
        return "InventorySlotPacket"

    def write(self, stream: BinaryStream) -> None:
        stream.write_unsigned_varint(self.container_id)
        stream.write_unsigned_varint(self.slot)
        self.container_name.write(stream)
        self.storage.write_header(stream)
        self.item.write(stream)

    def read(self, stream: ReadOnlyBinaryStream) -> None:
        pass



from collections import deque
from enum import Enum
from typing import TYPE_CHECKING

from bedrock_protocol.packets.types import BlockPos
from endstone import Player

from endstone_inventoryui.menu.graphic.block_graphic import BlockGraphic
from endstone_inventoryui.menu.graphic.block_pair_graphic import BlockPairGraphic
from endstone_inventoryui.menu.graphic.graphic import Graphic

if TYPE_CHECKING:
    from endstone_inventoryui.menu import Menu
from endstone_inventoryui.network.inventory_content_packet import InventoryContentPacket
from endstone_inventoryui.network.inventory_slot_packet import InventorySlotPacket
from endstone_inventoryui.network.item_stack_wrapper import ItemStackWrapper
from endstone_inventoryui.util.item_utils import is_air
from endstone_inventoryui.util.utils import send_ack_packet, get_block_behind


class Session:
    CONTAINER_ID: int = 200

    MAX_OPEN_ATTEMPTS: int = 10

    class State(Enum):
        NONE = 0
        GRAPHIC_SENT = 1
        GRAPHIC_RECEIVED = 2
        GRAPHIC_DATA_SENT = 3
        GRAPHIC_DATA_RECEIVED = 4
        OPENING = 5
        OPEN = 6
        CLOSING = 7

    def __init__(self, player: Player):
        self.player: Player = player
        self._menu: 'Menu | None' = None
        self.state: Session.State = self.State.NONE
        self.graphic: Graphic | None = None
        self.block_pos: list[BlockPos] = []
        self.open_attempts = 0
        self.ack_timestamp = 0
        self.pending: deque['Menu'] = deque()

    @property
    def menu(self) -> 'Menu | None':
        return self._menu

    @menu.setter
    def menu(self, value: 'Menu | None') -> None:
        if self._menu is not None:
            self._menu._remove_session(self)
        self._menu = value
        if value is not None:
            value._add_session(self)

    def send_menu(self):
        self.open_attempts = 0
        self.ack_timestamp = 0
        pos = get_block_behind(self.player, 2)
        self.graphic = BlockPairGraphic(self.menu, pos) if self.menu.type.is_pair else BlockGraphic(self.menu, pos)
        self.send_graphic()

    def send_graphic(self):
        self.graphic.send(self.player)
        self.state = self.State.GRAPHIC_SENT
        self.ack_timestamp = send_ack_packet(self.player)

    def send_graphic_data(self):
        self.graphic.send_data(self.player)
        self.state = self.State.GRAPHIC_DATA_SENT
        self.ack_timestamp = send_ack_packet(self.player)

    def open(self):
        self.state = self.State.OPENING
        self.graphic.open(self.player)
        self.ack_timestamp = send_ack_packet(self.player)

    def send_contents(self):
        inventory = self.menu.inventory
        pk = InventoryContentPacket(Session.CONTAINER_ID)
        for i in range(inventory.size):
            item_stack = inventory.get_item(i)
            pk.items.append(ItemStackWrapper(0 if is_air(item_stack) else i + 1, item_stack))

        self.player.send_packet(pk.get_packet_id(), pk.serialize())

    def update_slot(self, slot: int):
        item = self.menu.inventory.get_item(slot)
        pk = InventorySlotPacket(self.CONTAINER_ID, slot, item=ItemStackWrapper(0, item))
        self.player.send_packet(pk.get_packet_id(), pk.serialize())

    def close(self):
        self.state = self.State.CLOSING
        if self.graphic is not None:
            self.graphic.remove(self.player)

    def update_state(self, state: State):
        self.state = state
        match state:
            case self.State.GRAPHIC_RECEIVED:
                self.send_graphic_data()
            case self.State.GRAPHIC_DATA_RECEIVED:
                self.open()
            case self.State.OPEN:
                self.send_contents()

    def __del__(self):
        self.close()



from ctypes import c_int16

from bedrock_protocol.packets.types import ItemData
from endstone._internal.endstone_python import ItemStack, ItemMeta
from rapidnbt import CompoundTag, ListTag

enchant_map = {
    "protection": 0,
    "fire_protection": 1,
    "feather_falling": 2,
    "blast_protection": 3,
    "projectile_protection": 4,
    "thorns": 5,
    "respiration": 6,
    "depth_strider": 7,
    "aqua_affinity": 8,
    "sharpness": 9,
    "smite": 10,
    "bane_of_arthropods": 11,
    "knockback": 12,
    "fire_aspect": 13,
    "looting": 14,
    "efficiency": 15,
    "silk_touch": 16,
    "unbreaking": 17,
    "fortune": 18,
    "power": 19,
    "punch": 20,
    "flame": 21,
    "infinity": 22,
    "luck_of_the_sea": 23,
    "lure": 24,
    "frost_walker": 25,
    "mending": 26,
    "binding": 27,
    "vanishing": 28,
    "impaling": 29,
    "riptide": 30,
    "loyalty": 31,
    "channeling": 32,
    "multishot": 33,
    "piercing": 34,
    "quick_charge": 35,
    "soul_speed": 36,
    "swift_sneak": 37,
    "wind_burst": 38,
    "density": 39,
    "breach": 40,
}

cached_items: dict[str, ItemData] = {}


def is_air(item_stack: ItemStack) -> bool:
    return item_stack.type.id == "minecraft:air"


def clone_item(item_stack: ItemStack) -> ItemStack:
    new_item = ItemStack(item_stack.type.id, item_stack.amount, item_stack.data)
    new_item.set_item_meta(item_stack.item_meta.clone())
    return new_item


def all_item_data() -> dict[str, ItemData]:
    return cached_items


def add_item_data(item_id: str, data: ItemData) -> None:
    cached_items[item_id] = data


def get_item_data(item_id: str) -> ItemData | None:
    return cached_items.get(item_id)


def get_enchant_type(enchant: str):
    global enchant_map
    return enchant_map.get(enchant, 42)  # 42 = InvalidEnchantment


def build_tag(item_meta: ItemMeta) -> CompoundTag:
    """Build the NBT compound tag for item metadata"""
    tag = CompoundTag()
    if item_meta.has_display_name or item_meta.has_lore:
        display_tag = CompoundTag()
        if item_meta.has_display_name:
            display_tag.set("Name", item_meta.display_name)
        if item_meta.has_lore:
            lore_list = ListTag()
            for lore_line in item_meta.lore:
                lore_list.append(lore_line)
            display_tag.set("Lore", lore_list)
        tag.set("display", display_tag)

    if item_meta.has_enchants:
        ench_list = ListTag()
        for enchant, level in item_meta.enchants.items():
            ench_tag = CompoundTag()
            ench_tag.set("id", c_int16(get_enchant_type(enchant)))
            ench_tag.set("lvl", c_int16(level))
            ench_list.append(ench_tag)
        tag.set("ench", ench_list)

    if item_meta.has_repair_cost:
        tag.set("RepairCost", item_meta.repair_cost)

    if item_meta.is_unbreakable:
        tag.set("Unbreakable", 1)
    return tag
